
(* Modify only typecheck and interpret *)

open Ast

exception TypeError
exception RunTimeError
exception Unimplemented (* shows you what to change *)

(*
(* commented out, but an approach equivalent to the one below using
   explicit mutation
*)
(*********** environment implementation where env = var->exp **********)

(* Note: types do not matter at run-time, so we just use Ignore to fill 
   in for a type we do not care about *) 

let mt_env x = raise RunTimeError

let extend_env env x e = fun y -> if x=y then e else env y

(* call this from the Rec case of interpret *)
let extend_env_rec env f x e = 
  let rec new_env y =
    if y=f
    then Closure(Fun(x,Ignore,e), new_env)
    else env y
  in new_env

let lookup_env env y = env y
*)

(******* environment implementation where env = (var * (exp ref)) list *******)

(* Note: types do not matter at run-time, so we just use Ignore to fill 
   in for a type we do not care about *) 

let mt_env = []

let extend_env env x e = (x,ref e)::env

(* call this from the Rec case of interpret *)
let extend_env_rec env f x e =
  let r       = ref True in
  let new_env = (f,r)::env in
  let _       = r := Closure(Fun(x,Ignore,e),new_env) in
  new_env

let rec lookup_env env y =
  match env with
    [] -> raise RunTimeError
  | (x,r)::tl -> if x=y then !r else lookup_env tl y

(* type-checker *)
(* Note: t1=t2 will do what you want to compare types structurally, 
   e.g., Arrow(Int,Bool)=Arrow(Int,Bool)
   But t1==t2 is roughly pointer-equality -- do not accidentally use it or
   you will be very sorry *)

let mt_ctxt = []
let extend_ctxt ctxt x t = (x,t)::ctxt
let rec lookup_ctxt ctxt y =
  match ctxt with
    [] -> raise TypeError
  | (x,t)::tl -> if x=y then t else lookup_ctxt tl y

(* do not change anything above here *)

(* we even did two cases for you! *)
let rec typecheck ctxt e =
  match e with
    True  -> Bool
  | First(e1) ->
      (match typecheck ctxt e1 with
	Pair(t1,t2) -> t1
      |	_ -> raise TypeError)
  | _ -> raise Unimplemented

(* evaluator 
   remember we are large-step -- return a value
   *)
type version = PartC | PartD | PartE | PartF

(* we even did two cases for you! *)
let rec interpret v env e =
  let interp = interpret v in (* make recursive calls more convenient *)
  match e with
    True -> e
  | First(e1) ->
      (match interp env e1 with
	MakePair(x,y) -> x
      |	_ -> raise RunTimeError)
  | MakePair(e1,e2) -> 
      (match v with
	PartC -> raise Unimplemented
      |	PartD -> raise Unimplemented
      |	PartE -> raise Unimplemented
      |	PartF -> raise Unimplemented)
  | _ -> raise Unimplemented

(*Do not change anything below here*)

(* expect one command-line argument, a file to parse *)
(* you do not need to understand this interaction with the system *)
let parse_file () =
  let argv = Sys.argv in
  let _ = 
    if Array.length argv != 2
    then (prerr_string ("usage: " ^ argv.(0) ^ " [file-to-parse]\n");
	  exit 1) in
  let ch = open_in argv.(1) in
  Parse.program Lex.lexer (Lexing.from_channel ch)

let _ =
  let prog = parse_file() in
  (try
    (ignore(typecheck mt_ctxt prog);
(*     print_string (exp2string prog); *) (* useful for debugging *)
(* comment some of these out until you are done perhaps... *)
     print_string ("\ninterpreted answer = ");
     print_string (exp2string (interpret PartC mt_env prog));
     print_string ("\ninterpreted answer2 = ");
     print_string (exp2string (interpret PartE mt_env prog));
     print_string ("\ninterpreted answer3 = ");
     print_string (exp2string (interpret PartE mt_env prog));
     print_string ("\ninterpreted answer4 = ");
     print_string (exp2string (interpret PartF mt_env prog));
     print_newline ())
  with TypeError -> print_string "did not typecheck\n")
